#!/usr/bin/env python3
import os
import sys
import nbdkit
import builtins


"""pssd.py: Preserve SSD"""
__version__   = "1.0"
__author__    = "ed <pssd@ocv.me>"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2020

"""
  creates a virtual blockdevice; a proxy to a real blockdevice

  read-calls from the OS are let through just as usual

  write-calls from the OS are first compared to the existing data,
  discarding all writes where the data matches, reducing the SSD wear

  you know what they say! "use pssd to avoid eventually being pssd"

========================================================================

  this needs nbdkit:
    git clone https://github.com/libguestfs/nbdkit.git
    cd nbdkit && autoreconf -i && ./configure && make -j$(nproc)

  example usage:
    ./nbdkit -f python ~ed/dev/pssd.py file=/dev/sda
    nbd-client 127.0.0.1
    pixz -dtk < backup.dd.xz | pv > /dev/nbd0
    nbd-client -d /dev/nbd0

"""


API_VERSION = 2


src_path = None
def config(key, value):
    global src_path
    if key == "file":
        src_path = os.path.abspath(value)
    else:
        raise RuntimeError("unknown parameter: " + key)


def config_complete():
    global src_path
    if src_path is None:
        raise RuntimeError("file parameter is required")


# no effect unless there are sleeps inside callbacks
def thread_model():
    return nbdkit.THREAD_MODEL_PARALLEL


total = [0, 0, 0]
def report(*args):
    global total
    total = [x + y for x, y in zip(total, args)]
    print('\033[A{:.0f}M read, {:.0f}M skip, {:.0f}M write'.format(
        *[x / (1024 * 1024) for x in total]), file=sys.stderr) 


# called on connect
def open(readonly):
    if readonly:
        flags = os.O_RDONLY
    else:
        flags = os.O_RDWR
    
    fd = os.open(src_path, flags)
    print('connected', file=sys.stderr)
    return { 'fd': fd }


def get_size(h):
    #return os.stat(h['fd']).st_size
    with builtins.open(src_path, 'rb') as f:
        f.seek(0, os.SEEK_END)
        return f.tell()


def pread(h, buf, offset, flags):
    n = os.preadv(h['fd'], [buf], offset)
    report(n, 0, 0)
    if n != len(buf):
        raise RuntimeError("unexpected short read from file")


def pwrite(h, buf, offset, flags):
    fd = h['fd']
    cbuf = os.pread(fd, len(buf), offset)
    if buf == cbuf:
        return report(len(cbuf), len(cbuf), 0)
    
    nw = os.pwritev(fd, [buf], offset)
    report(len(cbuf), 0, nw)
    if nw != len(buf):
        raise RuntimeError("short write")

